from cnvrg.modules.library import Library
from cnvrg.modules.experiment import Experiment
from cnvrg.modules.endpoint import Endpoint
from cnvrg.modules.data_connector.mysql_connector import MysqlConnector
from cnvrg.version import VERSION



__version__ = VERSION


import os

EXPERIMENT = "Experiment"
ENDPOINT = "Endpoint"


POOL_SIZE = os.environ.get("CNVRG_THREAD_SIZE") or 20
CURRENT_JOB_ID = os.environ.get("CNVRG_JOB_ID")
CURRENT_JOB_TYPE = os.environ.get("CNVRG_JOB_TYPE")
CNVRG_OUTPUT_DIR = os.environ.get("CNVRG_OUTPUT_DIR")
MAX_LOGS_PER_SEND = int(os.environ.get("CNVRG_MAX_LOGS_PER_SEND") or 500)

def in_experiment():
    return CURRENT_JOB_TYPE == EXPERIMENT


def get_origin_job():
    if not CURRENT_JOB_ID and not CURRENT_JOB_TYPE: return {}
    return {
        "origin_job_id": CURRENT_JOB_ID,
        "origin_job_type": CURRENT_JOB_TYPE
    }


import cnvrg.helpers.apis_helper as apis_helper
import cnvrg.helpers.param_build_helper as param_build_helper

class BaseConnector():
    def __init__(self, data_connector):
        org, data_connector = param_build_helper.parse_params(data_connector, param_build_helper.DATA_CONNECTOR)
        self._org = org
        self._data_connector = data_connector
        self.data = self.__init_data_connector()

    def __init_data_connector(self):
        return apis_helper.get(self.__base_url()).get("data_connector")

    def __base_url(self):
        return apis_helper.url_join('users', self._org, "data_connectors", self._data_connector)


def __clean_start_end_slashes(s):
    if s.endswith("/"): s = s[:-1]
    if s.startswith("/"): s = s[1:]
    return s

def url_join(path, *paths):
    all_paths = [path] + list(paths)
    return "/".join(map(__clean_start_end_slashes, filter(lambda x: x, all_paths)))


from cnvrg.modules.cnvrg_files import CnvrgFiles
import cnvrg.helpers.param_build_helper as param_build_helper
import cnvrg.helpers.config_helper as config_helper
from cnvrg.modules.errors import CnvrgError
from cnvrg.modules.storage import storage_factory
import os.path

class Dataset(CnvrgFiles):
    def __init__(self, dataset=None, dataset_url=None, working_dir=None):
        if dataset_url:
            owner_slug, project_slug = Dataset.get_owner_and_project_from_url(dataset_url)
        else:
            owner_slug, project_slug = param_build_helper.parse_params(dataset, param_build_helper.DATASET, working_dir=working_dir)
        self.__owner = owner_slug
        self.__dataset = project_slug

        in_dir = config_helper.is_in_dir(config_helper.CONFIG_TYPE_DATASET, project_slug, working_dir)
        super(Dataset, self).__init__()
        self._set_working_dir(working_dir)
        if in_dir:
            self._set_working_dir(config_helper.find_config_dir(path=working_dir))
            self.in_dir = True
        elif not self.__dataset or not self.__owner:
            raise CnvrgError("Cant init project without params and outside project directory")

    def get_base_url(self):
        return "users/{owner}/datasets/{dataset}".format(owner=self.__owner, dataset=self.__dataset)

    def fetch_all_files(self):
        return self.get_commit_files()

    def pytorch_dataset(self, loader=None):
        from torch.utils.data.dataset import Dataset as TorchDataset
        class CnvrPytorchDataset(TorchDataset):
            'Characterizes a dataset for PyTorch'
            def __init__(self, dataset: CnvrgFiles=None):
                'Initialization'
                self.dataset = dataset
                self.files = dict(map(lambda x: [x.get("fullpath"), x], dataset.get_commit_files()))
                self.keys = list(self.files.keys())
                self.storage = storage_factory(dataset)

            def __len__(self):
                'Denotes the total number of samples'
                return len(self.files.keys())

            def __getitem__(self, index):
                'Generates one sample of data'
                # Select sample
                file_def = self.files[self.keys[index]]
                file_path = os.path.join(self.dataset.get_working_dir(), file_def.get("fullpath"))
                if not os.path.exists(file_path):
                    self.storage.download_single_file(file_def)
                return file_path if not loader else loader(file_path)


        return CnvrPytorchDataset(dataset=self)



from cnvrg.modules.errors import CnvrgError

def args_to_string(args):
    if args == None: return ''
    if isinstance(args, list): args = {k["key"]: k["value"] for k in args}
    ### expect dict of key=value
    return " ".join(map(lambda x: "--{key}='{value}'".format(key=x[0], value=x[1]), args.items()))


def validate_args(args):
    for key,value in args.items():
        if isinstance(value, list):
            raise CnvrgError("Library local run with many arguments are not supported yet.")


import hashlib
import os
import cnvrg.helpers.env_helper as env
import cnvrg.helpers.parallel_helper as parallel_helper
import cnvrg.helpers.cnvrgignore_helper as cnvrgignore_helper
from pathlib import Path
BUF_SIZE = 65536

def hash_dir(directory, pool_size=env.POOL_SIZE):
    files = list(map(str, Path(directory).rglob("*")))
    ignored = set(list(cnvrgignore_helper.get_full_ignore(directory)))
    files = set(files) - set(ignored)
    return dict(map(lambda x: [os.path.relpath(x[0], directory), x[1]], parallel_helper.safe_parallel(hash_file, filter(lambda x: not os.path.isdir(x), files), pool_size)))

def hash_file(file, raw=False):
    sha1 = hashlib.sha1()
    with open(file, 'rb') as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            sha1.update(data)
    if raw: return sha1.hexdigest()
    return str(file), {"sha1": sha1.hexdigest()}